package com.winhc.max.compute.graph.job.enterprise_group;

import com.aliyun.odps.graph.ComputeContext;
import com.aliyun.odps.graph.Edge;
import com.aliyun.odps.graph.Vertex;
import com.aliyun.odps.graph.WorkerContext;
import com.aliyun.odps.io.BooleanWritable;
import com.aliyun.odps.io.LongWritable;
import com.aliyun.odps.io.NullWritable;
import com.aliyun.odps.io.Text;
import com.aliyun.odps.utils.StringUtils;
import com.google.gson.Gson;
import com.winhc.max.compute.graph.job.enterprise_group.entity.EntGroupMsg;
import com.winhc.max.compute.graph.job.enterprise_group.entity.EnterpriseGroupVertexValue;
import com.winhc.max.compute.graph.job.enterprise_group.entity.HolderEdge;
import com.winhc.max.compute.graph.job.enterprise_group.entity.VertexComputeInfo;
import com.winhc.max.compute.graph.util.WritableRecordExtensions;
import lombok.experimental.ExtensionMethod;
import lombok.extern.slf4j.Slf4j;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author: XuJiakai
 * 2023/2/7 13:59
 */
@ExtensionMethod({
        WritableRecordExtensions.class
        , EnterpriseGroupUtils.class
})
@Slf4j
public class EnterpriseGroupVertex extends
        Vertex<Text, EnterpriseGroupVertexValue, HolderEdge, EntGroupMsg> {

    private static Gson gson;
    private static ProdEnvDebug prodEnvDebug;

    @Override
    public void setup(WorkerContext<Text, EnterpriseGroupVertexValue, HolderEdge, EntGroupMsg> context) throws IOException {
        gson = new Gson();
        prodEnvDebug = ProdEnvDebug.build(context.getConfiguration());
    }

    @Override
    public void compute(ComputeContext<Text, EnterpriseGroupVertexValue, HolderEdge, EntGroupMsg> context, Iterable<EntGroupMsg> messages) throws IOException {
        context.aggregate(NullWritable.get());
        if (getId().toString().length() == 33) {
            voteToHalt();
            //跳过人员节点
            return;
        }

        List<String> legalEntityIds = getValue().getLegalEntityIds().tuple2List();


        //获取需要传递的下一个节点，如无，当前企业则为控股控股公司。
        List<Edge<Text, HolderEdge>> edges = getEdges();
        VertexComputeInfo vertexInfo = VertexComputeInfo.getVertexInfo(edges, legalEntityIds, getValue().getManualStop().get());


        //todo:debug
        prodEnvDebug.debug_1(getId()
                , "\nvertexInfo: {}\n"
                , vertexInfo
        );

        //将本节点传递到下一节点
        if (!getValue().isSendMsg()) {
            for (String destVertexId : vertexInfo.getHoldCompanyId()) {
                //传递控股企业
                EntGroupMsg entGroupMsg = EntGroupMsg.ofByType_1(getId(), getValue().getCompanyName(), getValue().getCompanyType(), vertexInfo.getNotHoldKeyno());
                boolean flag = entGroupMsg.routeLog(destVertexId);
                if (!flag) {
//                    getValue().setOutput();
                    continue;
                }

                entGroupMsg.addStockRightControlChain(getId().toString(), vertexInfo.getInvestmentProportion(destVertexId), destVertexId);

                context.sendMessage(new Text(destVertexId), entGroupMsg);
                context.aggregate(new BooleanWritable(true));
            }

            for (String destVertexId : vertexInfo.getNotHoldCompanyId()) {
                //传递参股企业
                EntGroupMsg entGroupMsg = EntGroupMsg.ofByType_2(getId(), getValue().getCompanyName());
                boolean flag = entGroupMsg.routeLog(destVertexId);
                if (!flag) {
//                    getValue().setOutput();   todo:参股企业没必要输出
                    continue;
                }
                context.sendMessage(new Text(destVertexId), entGroupMsg);
                context.aggregate(new BooleanWritable(true));
            }

            if (vertexInfo.getHoldCompanyId().isEmpty()) {
                //如果没有控股企业，则本节点为集团，需将本节点的参股股东，添加至集团股东
                List<String> groupHolderCompanyIds = getValue().getGroupHolderCompanyIds().tuple2List();
                groupHolderCompanyIds.addAll(vertexInfo.getNotHoldKeyno());
                getValue().setGroupHolderCompanyIds(groupHolderCompanyIds);
            }

            getValue().setEndFlag();
        }


        //是否将消息汇总至该节点
        boolean collect = vertexInfo.getHoldCompanyId().isEmpty();

        if (collect) {
            //无法传递，则在本节点汇总
            mergeMsg2Vertex(getValue(), messages, vertexInfo.getControllerKeyno());
        } else {
            for (EntGroupMsg message : messages) {
                //将本节点获取的消息直接传递至下一节点，只传递到控股企业。不必将接收到的参股企业做区分，直接只往控股企业传递就行
                for (String destVertexId : vertexInfo.getHoldCompanyId()) {
                    boolean flag = message.routeLog(destVertexId);
                    if (!flag) {
                        //todo:debug
                       /*prodEnvDebug.debug_3(getId()
                                , "\nvertexInfo: {}\nroteLog:{}\n"
                                , vertexInfo
                                ,message
                        );
                        mergeMsg2Vertex(getValue(), Arrays.asList(message), null);*/
                        continue;
                    }
                    message.addStockRightControlChain(getId().toString(), vertexInfo.getInvestmentProportion(destVertexId), destVertexId);

                    context.sendMessage(new Text(destVertexId), message);
                    context.aggregate(new BooleanWritable(true));
                }
            }
        }

        voteToHalt();
    }

    /**
     * 将msg信息合并到当前节点
     *
     * @param vertexEntity
     * @param messages
     * @param controllerKeyno
     */
    private static void mergeMsg2Vertex(EnterpriseGroupVertexValue vertexEntity, Iterable<EntGroupMsg> messages, String controllerKeyno) {
        long holdNum = vertexEntity.getHoldNum().get();
        List<String> holdCompanyId = vertexEntity.getHoldCompanyIds().tuple2List();
        List<String> groupInvestmentCompanyId = vertexEntity.getGroupInvestmentCompanyIds().tuple2List();
        List<String> groupHolderCompanyId = vertexEntity.getGroupHolderCompanyIds().tuple2List();
        List<String> subgroup = vertexEntity.getSubgroup().tuple2List();
        List<String> stockRightControlChain = Arrays.stream(vertexEntity.getStockRightControlChain().toString().split(",")).collect(Collectors.toList());

        if (subgroup == null) {
            subgroup = new ArrayList<>();
        }

        for (EntGroupMsg message : messages) {
            holdNum += message.getHoldNum().get();
            holdCompanyId.addAll(message.getHoldCompanyIds().tuple2List());
            groupHolderCompanyId.addAll(message.getGroupHolderCompanyIds().tuple2List());
            groupInvestmentCompanyId.addAll(message.getGroupInvestmentCompanyIds().tuple2List());
            stockRightControlChain.add(message.getStockRightControlChain().toString());
            if (message.getSourceVertexId() != null && StringUtils.isNotBlank(message.getSourceVertexId().toString())) {
                subgroup.add(message.getSourceVertexId().toString());
            }
        }
        Text groupControllerKeyno;
        if (controllerKeyno != null) {
            groupControllerKeyno = new Text(controllerKeyno);
        } else {
            groupControllerKeyno = new Text();
        }

        String collect = stockRightControlChain.stream().filter(StringUtils::isNotBlank).collect(Collectors.joining(","));

        vertexEntity.setHoldNum(holdNum);
        vertexEntity.setHoldCompanyIds(holdCompanyId);
        vertexEntity.setGroupInvestmentCompanyIds(groupInvestmentCompanyId);
        vertexEntity.setGroupHolderCompanyIds(groupHolderCompanyId);
        vertexEntity.setGroupControllerKeyno(groupControllerKeyno);
        vertexEntity.setStockRightControlChain(new Text(collect));
        vertexEntity.setSubgroup(subgroup.list2Tuple());
        vertexEntity.setOutput();
    }

    @Override
    public void cleanup(WorkerContext<Text, EnterpriseGroupVertexValue, HolderEdge, EntGroupMsg> context) throws IOException {
        if (getId().toString().length() == 33) {
            //跳过人员节点
            return;
        }

        //todo:debug
        prodEnvDebug.debug_2(getId()
                , "vertex value:\n{}"
                , getValue()
        );

        if (!getValue().isOutput()) {
            return;
        }

        //输出
        Text enterprise_group_hold_company_id = getId();
        Text enterprise_group_hold_company_name = getValue().getCompanyName();

        Set<String> holdCompanyId = getValue().getHoldCompanyIds().tuple2Set();
        Text group_company_id = new Text(gson.toJson(holdCompanyId));

        //todo：可能会出现集团股东 本就是控股企业，需要排除
        Set<String> groupInvestmentCompanyId = getValue().getGroupInvestmentCompanyIds().tuple2Set();
        groupInvestmentCompanyId.removeAll(holdCompanyId);
        Text group_investment_company_id = new Text(gson.toJson(groupInvestmentCompanyId));

        Set<String> groupHolderCompanyId = getValue().getGroupHolderCompanyIds().tuple2Set();
        groupHolderCompanyId.removeAll(holdCompanyId);
        Text group_holder_company_id = new Text(gson.toJson(groupHolderCompanyId));


        Text group_controller_keyno = getValue().getGroupControllerKeyno();
        Text stockRightControlChain = getValue().getStockRightControlChain();

        LongWritable hold_num = new LongWritable(holdCompanyId.size() + 1);


        List<String> subgroupList = getValue().getSubgroup().tuple2ListByDistinct();
        Text subgroup = new Text();
        if (getValue().getCompanyName().toString().contains("集团")) {
            subgroup = new Text(subgroupList.stream().filter(e -> !e.equals(getId().toString())).collect(Collectors.joining(",")));
        }


        if (hold_num.get() < 3) {
            return;
        }
        if (hold_num.get() > 200) {
            String s = EnterpriseGroupUtils.stockChainTrim(enterprise_group_hold_company_id.toString(), stockRightControlChain.toString(), 10);
            stockRightControlChain = new Text(s);
        }


        //合并输出
        context.write(
                "parent"
                , enterprise_group_hold_company_id
                , enterprise_group_hold_company_id
                , enterprise_group_hold_company_name
                , hold_num
                , group_company_id
                , group_investment_company_id
                , group_holder_company_id
                , group_controller_keyno
                , stockRightControlChain
                , subgroup
        );

        //分批输出
        context.write(
                "main"
                , enterprise_group_hold_company_id
                , enterprise_group_hold_company_id
                , enterprise_group_hold_company_name
                , hold_num
                , group_controller_keyno
                , stockRightControlChain
                , subgroup
        );

        context.write(
                "relation"
                , new Text(enterprise_group_hold_company_id + "_0")
                , enterprise_group_hold_company_id
                , new Text("0")
                , enterprise_group_hold_company_id
        );

        for (String e : holdCompanyId) {
            context.write(
                    "relation"
                    , new Text(enterprise_group_hold_company_id + "_1_" + e)
                    , enterprise_group_hold_company_id
                    , new Text("1")
                    , new Text(e)
            );
        }

        for (String e : groupInvestmentCompanyId) {
            context.write(
                    "relation"
                    , new Text(enterprise_group_hold_company_id + "_2_" + e)
                    , enterprise_group_hold_company_id
                    , new Text("2")
                    , new Text(e)
            );
        }

        for (String e : groupHolderCompanyId) {
            if(e.length()==32){
                context.write(
                        "relation"
                        , new Text(enterprise_group_hold_company_id + "_3_" + e)
                        , enterprise_group_hold_company_id
                        , new Text("3")
                        , new Text(e)
                );
            }else{
                context.write(
                        "relation"
                        , new Text(enterprise_group_hold_company_id + "_4_" + e)
                        , enterprise_group_hold_company_id
                        , new Text("4")
                        , new Text(e)
                );
            }
        }

    }
}
