package com.winhc.max.compute.graph.job.enterprise_group.entity;

import com.aliyun.odps.graph.Edge;
import com.aliyun.odps.io.Text;
import com.winhc.max.compute.graph.job.enterprise_group.EnterpriseGroupUtils;
import com.winhc.max.compute.graph.util.WritableRecordExtensions;
import lombok.experimental.ExtensionMethod;

import java.util.*;
import java.util.stream.Collectors;

/**
 * @author: XuJiakai
 * 2023/3/6 10:33
 * <p>
 * 图计算节点信息处理
 */
@ExtensionMethod({
        WritableRecordExtensions.class
        , EnterpriseGroupUtils.class
})
public class VertexComputeInfo {

    /**
     * 该节点疑似实际控制人，多个用逗号隔开
     * 因为存在国务院等机关单位，所以该id可能是公司
     */
    private String controllerKeyno;

    /**
     * 控股公司（本质是控股公司子集，因需要过滤机关单位和不存在节点）
     * 用于向控股公司传递消息
     */
    private List<String> holdCompanyId = new ArrayList<>();


    /**
     * 非控股公司
     * 用于向非控股公司传递消息
     * 用于集团股东
     */
    private List<String> notHoldCompanyId = new ArrayList<>();


    /**
     * 非控股的人
     * 用于集团股东
     */
    private List<String> notHoldPersonId = new ArrayList<>();

    /**
     * 手动拆分集团标志位，true为手动拆分
     */
    private boolean manualStop;


    private Map<String, Double> majorityShareholder = new HashMap<>();


    public static VertexComputeInfo getVertexInfo(List<Edge<Text, HolderEdge>> edges, List<String> legalEntityIds, boolean manualStop) {
        if (edges == null || edges.isEmpty()) {
            VertexComputeInfo vertexComputeInfo = new VertexComputeInfo();
            vertexComputeInfo.manualStop = manualStop;
            return vertexComputeInfo;
        }


        //控股股东，及出资比例
        Map<String, Double> majorityShareholder = edges.getMajorityShareholder(legalEntityIds);
        /**
         * 计算控股股东，用于消息发送
         * step 1:如果需要手动拆分，则没有控股股东，和参股股东
         * step 2:过滤国有企业
         * step 3:过滤节点不存在公司，防止集团数据丢失
         */
        if (manualStop) {
            VertexComputeInfo vertexComputeInfo = new VertexComputeInfo();
            vertexComputeInfo.manualStop = true;
            vertexComputeInfo.majorityShareholder = majorityShareholder;
            vertexComputeInfo.controllerKeyno = String.join(",", majorityShareholder.keySet());
            return vertexComputeInfo;
        }

        //国有企业id或节点不存在公司id
        Set<String> filterCompanyId = edges.stream().filter(e -> {
            int i = e.getValue().getHolderType().get();
            return i == 2 || i == 4;
        }).map(e -> e.getDestVertexId().toString())
                .filter(e -> e.length() == 32)
                .collect(Collectors.toSet());

        //控股公司id，不包含个人
        List<String> thisVertexHoldCompanyId = majorityShareholder.keySet()
                .stream()
                .filter(e -> e.length() == 32)
                .filter(e -> !filterCompanyId.contains(e))
                .collect(Collectors.toList());


        //参股公司id，不包含个人
        List<String> thisVertexOtherHolderCompanyId = edges.getOtherHolder(majorityShareholder.keySet())
                .stream().filter(e -> e.length() == 32)
                .collect(Collectors.toList());

        //参股公司id，不包含个人
        List<String> thisVertexOtherHolderPersonId = edges.getOtherHolder(majorityShareholder.keySet())
                .stream().filter(e -> e.length() == 33)
                .collect(Collectors.toList());

        String controllerKeyno = String.join(",", majorityShareholder.keySet());

        VertexComputeInfo vertexComputeInfo = new VertexComputeInfo();
        vertexComputeInfo.manualStop = false;

        vertexComputeInfo.holdCompanyId = thisVertexHoldCompanyId;
        vertexComputeInfo.notHoldCompanyId = thisVertexOtherHolderCompanyId;
        vertexComputeInfo.notHoldPersonId = thisVertexOtherHolderPersonId;
        vertexComputeInfo.majorityShareholder = majorityShareholder;
        vertexComputeInfo.controllerKeyno = controllerKeyno;

        return vertexComputeInfo;
    }


    public List<String> getNotHoldPersonId() {
        return notHoldPersonId;
    }

    public List<String> getNotHoldKeyno(){
        List<String> result = new ArrayList<String>();
        result.addAll(notHoldPersonId);
        result.addAll(notHoldCompanyId);
        return result;
    }

    public String getControllerKeyno() {
        return controllerKeyno;
    }

    public List<String> getHoldCompanyId() {
        return holdCompanyId;
    }

    public List<String> getNotHoldCompanyId() {
        return notHoldCompanyId;
    }

    public boolean isManualStop() {
        return manualStop;
    }

    public Double getInvestmentProportion(String keyno) {
        return majorityShareholder.get(keyno);
    }

    @Override
    public String toString() {
        return "VertexComputeInfo{" +
                "controllerKeyno='" + controllerKeyno + '\'' +
                ", holdCompanyId=" + holdCompanyId +
                ", notHoldCompanyId=" + notHoldCompanyId +
                ", manualStop=" + manualStop +
                '}';
    }
}
